<?php
// $Id: XoxoExportableFactory.php,v 1.1.4.5 2010/12/27 19:43:45 pounard Exp $

/**
 * Defines basis for any kind of complex objets. 
 */
class XoxoExportableFactory extends XoxoFactory
{
  /**
   * @var array
   */
  protected $_loadedNames = array();

  /**
   * @var array
   */
  protected $_loadedNamesByOid = array();

  /**
   * Override the default constructor. This in order to save defaults as objects
   * in database for paging queries to work as any others.
   */
  public function __construct(&$item) {
    parent::__construct($item);
    // Check for defaults cache.
    if (!variable_get('xoxo_defaults_loaded', FALSE)) {
      // $this->saveDefaults();
    }
  }

  /**
   * Particular implementation that is able to use the object unique name as
   * caching parameter.
   * 
   * @see XoxoFactory::_cacheAdd()
   */
  public function _cacheAdd(XoxoObject $object, $toBinary = FALSE) {
    $oid = $object->getIdentifier();
    $name = $object->getName();

    // Parent cache (really important).
    $this->_loaded[$oid] = $object;
    // Our custom cache.
    $this->_loadedNames[$name] = $object;
    $this->_loadedNamesByOid[$oid] = $name;

    // Binary cache.
    if ($toBinary) {
      // FIXME:: Ugly cache duplication. cache_get() function does not seems to
      // be able to fetch using wildcards.
      cache_set($this->getTableName() . '-' . $oid, $object, XOXO_CACHE_OBJECT_TABLE);
      cache_set($this->getTableName() . '-' . $name, $object, XOXO_CACHE_OBJECT_TABLE);
    }
  }

  /**
   * Particular implementation that is able to use the object unique name as
   * caching parameter.
   * 
   * @see XoxoFactory::_cacheGetBinary()
   */
  protected function _cacheGetBinary() {
    $args = func_get_args();

    $oid = isset($args[0]) ? $args[0] : NULL;
    $name = isset($args[1]) ? $args[1] : NULL;

    if ($oid) {
      $cid = $this->getTableName() . '-' . $oid;
    }
    else if ($name) {
      $cid = $this->getTableName() . '-' . $name;
    }

    if ($data = cache_get($cid, XOXO_CACHE_OBJECT_TABLE)) {
      return $data->data;
    }
  }

  /**
   * Get all defaults.
   */
  public function saveDefaults() {
    $hook = $this->_type . '_defaults';
    foreach (module_implements($hook) as $module) {
      foreach (module_invoke($module, $hook) as $object) {
        try {
          // For each default, check if an override exists, if not, save it into
          // database.
          if (!$this->existsByName($object->getName())) {
            // Explicitely set the default boolean.
            $object->setDefault(TRUE);
            // The save() method will update internal static cache, so there is
            // nothing else to care about here.
            $this->save($object);
          }
        }
        catch (OoxException $e) {
          // Any exception here tells that the default definition is incomplete
          // so we just ignore it, but warn the developer he does an error.
          watchdog('xoxo', "A default object defined by module " . $module . " has a wrong definition (check if name is set), ignoring it.", NULL, WATCHDOG_ERROR);
        }
      }
    }
  }

  /**
   * (non-PHPdoc)
   * @see XoxoFactory::_populate()
   */
  protected function _populate($row, XoxoObject $object) {
    parent::_populate($row, $object);
    if ($object instanceof XoxoExportableObject) {
      $object->setName($row->name);
    }
  }

  /**
   * Load layout by machine name.
   * 
   * @param string $name
   *   Machine name.
   * 
   * @return XoxoExportableObject
   *   XoxoExportableObject instance.
   * 
   * @throws XoxoFactoryException
   *   If object does not exists.
   */
  public function loadByName($name) {
    // Attempt cache fetch.
    if (!isset($this->_loadedNames[$name])) {
      $object = NULL;

      // Attempt from cache first time.
      if ($object = $this->_cacheGetBinary(NULL, $name)) {
        $this->_cacheAdd($object, FALSE);
      }

      // Load from database aptempt.
      else {
        $data = db_select($this->getTableName(), 'ct')
          ->fields('ct')
          ->condition('ct.name', $name)
          ->execute()
          ->fetchObject();

        if ($data) {
          // This will populate the new object.
          $object = $this->getItemRegistry()->getItem($data->type);
          $this->_populate($data, $object);
          $this->_cacheAdd($object, TRUE);
        }

        // Fallback, object does not exists.
        else {
          throw new XoxoFactoryException("Object '" . $name . "' does not exists.");
        }
      }
    }

    $ret = NULL;

    // Apply cloning method.
    switch ($this->_loadMethod) {
      case self::LOAD_METHOD_CLONE:
        $ret = clone($this->_loadedNames[$name]);
        break;
      case self::LOAD_METHOD_INSTANCE:
      default:
        $ret = $this->_loadedNames[$name];
    }

    return $ret;
  }

  /**
   * Tells if current layout name exists as a default layout.
   * 
   * @return boolean
   */
  public function existsAsDefault($name) {
    // FIXME: This is kind of ugly, and could be optimized.
    $hook = $this->_type . '_defaults';
    foreach (module_implements($hook) as $module) {
      foreach (module_invoke($module, $hook) as $object) {
        if ($object->getName() == $name) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * (non-PHPdoc)
   * @see XoxoFactory::_preSave()
   * 
   * Last chance object save alteration, set our own real object name.
   */
  protected function _preSave(XoxoObject $object, &$data) {
    parent::_preSave($object, $data);
    $data['name'] = $object->getName();
  }

  /**
   * Delete given layout.
   * 
   * If layout does not exists, this method will remain silent.
   * 
   * @param string $name
   *   XoxoExportableObject name.
   */
  public function deleteByName($name) {
    return $this->deleteByField('name', $name);
  }

  /**
   * Check if given machine name already exists in database.
   * 
   * @param string $name
   *   Machine name.
   * 
   * @return boolean
   */
  public function existsByName($name) {
    return (bool) db_query("SELECT 1 FROM {:table} WHERE name = :name", array(
      ':table' => $this->getTableName(),
      ':name' => $name
    ))->fetchField();
  }

  /**
   * Import layout from raw PHP code.
   * In case of any error, no records at all will be saved.
   * 
   * Be careful, if any of these layouts exists, it will overwrite the
   * existing one.
   * 
   * Note for developers: when you override this method, be careful about the
   * $pretend parameter, and dont forget to call the XoxoFactory::save()
   * method in order to save new objects.
   * 
   * @param string $code
   *   Raw PHP code returning a XoxoExportableObject instance, or an array of
   *   XoxoExportableObject instances.
   * @param boolean $pretend = FALSE
   *   Set this to TRUE if you don't want code to be actually imported. This
   *   can be usefull for forms validation phase.
   * 
   * @throws XoxoFactoryException
   *   In case of any error. Error message will be a localized human readable
   *   error message and will be displayed to end use in the import form.
   * 
   * @return array
   *   The array of new imported layouts.
   */
  public function importCode($code, $pretend = FALSE) {
    throw new XoxoFactoryException("Not implemented");
  }

  /**
   * Export the given object to raw PHP code.
   * 
   * This is the method that specific factories should override. This simple
   * implementation will be enougth for objects using only options as storage.
   * 
   * @param XoxoExportableObject $object
   *   An XoxoExportableObject object specific instance.
   * 
   * @return array
   *   An array of PHP code lines.
   * 
   * @throws XoxoFactoryException
   *   In case any error happens.
   */
  public function exportSingle(XoxoExportableObject $object, $objectName = 'object') {
    $code = array();

    // Set common properties, exlude the oid..
    $code[] = "\$" . $objectName . " = new " . $this->_className . "();";
    $code[] = "\$" . $objectName . "->setStatus(" . $object->getStatus() .  ");";
    $code[] = "\$" . $objectName . "->setOwner(" . $object->getOwnerUid() .  ");";
    $code[] = "\$" . $objectName . "->setName('" . stripslashes($object->getName()) .  "');";
    if ($module = $object->getModule()) {
      $code[] = "\$" . $objectName . "->setModule('" . stripslashes($module) .  "');";
    }
    if ($description = $object->getDescription()) {
      $code[] = "\$" . $objectName . "->setDescription('" . stripslashes($description) .  "');";
    }

    foreach ($object->getOptions() as $name => $value) {
      if (is_string($value)) {
        $code[] = "\$" . $objectName . "->setOption('" . $name . "', '" . stripslashes($value) .  "');";
      }
      else if (is_numeric($value)) {
        $code[] = "\$" . $objectName . "->setOption('" . $name . "', " . $value .  ");";
      }
      else {
        $code[] = "\$" . $objectName . "->setOption('" . $name . "', " . serialize($value) .  ");";
      }
    }

    return $code;
  }

  /**
   * Export the given object(s) to raw PHP code.
   * 
   * @param array $object
   *   Array of XoxoExportableObject object instances.
   * @param boolean $returnStatement = TRUE
   *   (optional) Append or not a return statement at the end of the code.
   * @param int $indent = 0
   *   (optional) Number of space to indent the code.
   * 
   * @return array
   *   An array of PHP line codes.
   * 
   * @throws XoxoFactoryException
   *   In case any error happens.
   */
  public final function exportMutiple($objects, $returnStatement = TRUE, $indent = 0) {
    $code = array();
    $code[] = "\$items = array();";
    $code[] = "";

    foreach ($objects as $object) {
      $code = array_merge($code, $this->exportSingle($object, 'object'));
      $code[] = "\$items[] = \$object;";
      $code[] = "";
    }

    if ($returnStatement) {
      $code[] = "return \$items;";
    }

    return $code;
  }

  /**
   * Allow sub modules to expose new dependencies for the feature module
   * xoxo.
   * 
   * This is absolutely not mandatory, but if you intend to use the import
   * and export facilities, you are strongly adviced to implement this.
   * 
   * @see hook_features_export()
   * 
   * @param XoxoExportableObject $object
   *   The current object being exported
   * @param array &$pipe
   *   See features module documentation.
   * @param array &$export
   *   See features module documentation.
   * 
   * @throws XoxoFactoryException
   *   In case any error happens.
   */
  public function featuresDependencies($object, &$pipe, &$export) { }

  /**
   * (non-PHPdoc)
   * @see XoxoFactory::_cacheClear()
   */
  protected function _cacheClear($oids = array()) {
    parent::_cacheClear($oids);
    // An optimized implementation would do it in a single foreach() by fully
    // overriding the parent method. Still keeping it like this for other
    // implementors to have a sample of how to do it without breaking parent
    // implementation.
    if (!empty($oids)) {
      foreach ($oids as $oid) {
        unset($this->_loadedNames[$this->_loadedNamesByOid[$oid]]);
        unset($this->_loadedNamesByOid[$oid]);
      }
    }
    else {
      $this->_loadedNames = array();
      $this->_loadedNamesByOid = array();
    }
  }
}
