<?php
// $Id: XoxoFactory.php,v 1.1.4.9 2010/12/27 19:43:45 pounard Exp $

class XoxoFactoryException extends OoxException { }

/**
 * Default factory for XoxoObject instances.
 * 
 * The factory is able to create, read, update or insert any objects of the
 * factory target type, using transactionnal requests for create and update
 * operations.
 * Target type is a class name, given by the hook_xoxo_factory() module
 * custom implementation that defines it.
 * 
 * Factories are themselves registrered using the XoxoFactoryRegistry
 * specific OoxRegistry implementation. This is masked by some helpers
 * in the XoXo module such as xoxo_factory_get().
 * 
 * @see oox_factory_get()
 * @see XoxoFactoryRegistry
 * @see OoxRegistry
 */
class XoxoFactory
{
  // An error happend during save.
  const SAVED_ERROR = 0;
  // Object has been updated.
  const SAVED_UPDATED = 1;
  // Object has been saved.
  const SAVED_NEW = 2;

  // Clone object when loading it from cache.
  const LOAD_METHOD_CLONE = 1;
  // Get instance when loading it from cache.
  const LOAD_METHOD_INSTANCE = 2;

  // FIXME: Implement this.
  // Do not cache objects.
  const CACHE_METHOD_NONE = 0;
  // Cache object on a per-object basis.
  const CACHE_METHOD_OBJECT = 1;
  // Cache all known objects in one single cache entry.
  const CACHE_METHOD_ALL = 2;

  /**
   * Internal type name, usefull for the registry pattern.
   * 
   * @var string
   */
  private $__type = NULL;

  /**
   * Get internal type name.
   * 
   * @return string
   */
  public function getType() {
    return $this->__type;
  }

  /**
   * Item registry type.
   * 
   * @var string
   */
  private $__itemRegistryType;

  /**
   * Get the registry that manage this factory's objects instances.
   * 
   * @return OoxRegistry
   */
  public final function getItemRegistry() {
    return OoxRegistry::getInstance($this->__itemRegistryType);
  }

  /**
   * For path generation, Drupal path prefix for factory objects handling.
   */
  private $__path;

  /**
   * For path generation, get Drupal path prefix for factory objects handling.
   * 
   * @return string
   *   Drupal path prefix.
   */
  public final function getPath() {
    return $this->__path;
  }

  /**
   * SQL table name.
   * 
   * @var string
   */
  private $__tableName;

  /**
   * Get table name.
   * 
   * Override this method if you don't define the table name at hook level.
   * 
   * @return string
   *   Table name.
   */
  public function getTableName() {
    if (!isset($this->__tableName)) {
      throw new XoxoFactoryException("No table name set.");
    }
    return $this->__tableName;
  }

  /**
   * Tells if target class is exportable.
   * 
   * FIXME: really fix me, this is important.
   */
  public final function isExportable() {
    // return is_subclass_of($this->_className, 'XoxoExportableObject');
    return FALSE;
  }

  /**
   * Cache method.
   * 
   * @var int
   */
  protected $_cacheMethod = self::CACHE_METHOD_NONE;

  /**
   * Load method.
   * 
   * @var int
   */
  protected $_loadMethod = self::LOAD_METHOD_INSTANCE;

  /**
   * Object type machine name.
   * 
   * @var string
   */
  protected $_type;

  /**
   * Construct the factory using the given table name.
   *
   * @param array $item
   *   Descriptive registry type array.
   */
  public function __construct(&$item) {
    $this->__tableName = $item['table'];
    $this->_type = $item['type'];
    $this->__itemRegistryType = isset($item['registry_type']) ? $item['registry_type'] : $item['type'];
    if ($item['load_method']) {
      $this->_loadMethod = $item['load_method'];
    }
    $this->__path = isset($item['path']) ? $item['path'] : 'admin/objects/' . $item['type'];
  }

  /**
   * Maintain a loaded object cache.
   * 
   * @var array
   */
  protected $_loaded = array();

  /**
   * Populate new instance using given database row.
   * 
   * Override this method to add your own fields you may have added. Don't ever
   * forget to call the parent implementation.
   * 
   * @param object $row
   *   Full row from database.
   * @param XoxoObject $object
   *   New object from given row data.
   */
  protected function _populate($row, XoxoObject $object) {
    // Override some parameters that could have been saved with database
    // specified data. This will allow modules to alter data in database
    // without having to load/save the full object and ensure the real
    // loaded instance will keep these settings.
    $object->setIdentifier($row->oid);
    $object->setOwner($row->uid);
    $object->setStatus($row->status);
    $object->setDefault($row->defaults);
    // Load object options, if any.
    if ($row->options_oid) {
      OptionsFactory::getInstance()->load($row->options_oid, $object);
    }
    $object->_setFactory($this);
    $object->restore();
  }

  /**
   * Load an object by identifier.
   * 
   * Even if you load multiple instances, use a complex query to fetch only
   * identifiers then call this one to fetch objects one by one. By doing this
   * you will ensure that the cache mecanism is used.
   * 
   * Cache mecanism performance gain is almost nothing on basic objects. But as
   * soon as you will have your objects volumetry growing, and an options table
   * also growing, this will start to be really valuable.
   * 
   * @param int $oid
   *   Identifier.
   * 
   * @return XoxoObject
   *   XoxoObject instance.
   * 
   * @throws XoxoFactoryException
   *   If layout does not exists.
   */
  public final function loadByIdentifier($oid) {
    // Attempt cache fetch.
    if (!isset($this->_loaded[$oid])) {
      $object = NULL;

      // Attempt from cache first time.
      if ($object = $this->_cacheGetBinary($oid)) {
        $this->_cacheAdd($object, FALSE);
      }

      // Load from database aptempt.
      else {
        $data = db_select($this->getTableName(), 'ct')
          ->fields('ct')
          ->condition('oid', $oid)
          ->execute()
          ->fetchObject();

        if ($data) {
          // This will populate the new object.
          $object = $this->getItemRegistry()->getItem($data->type);
          $this->_populate($data, $object);
          $this->_cacheAdd($object, TRUE);
        }

        // Fallback, object does not exists.
        else {
          throw new XoxoFactoryException("Object '" . $oid . "' does not exists.");
        }
      }
    }

    $ret = NULL;

    // Apply cloning method.
    switch ($this->_loadMethod) {
      case self::LOAD_METHOD_CLONE:
        $ret = clone($this->_loaded[$oid]);
        break;
      case self::LOAD_METHOD_INSTANCE:
      default:
        $ret = $this->_loaded[$oid];
    }

    return $ret;
  }

  /**
   * Load multiple entries using the given identifiers. This is mainly an
   * optimization method.
   * 
   * Please use this method whenever you can.
   * 
   * @param array $identifiers
   *   Array of object unique identifiers. You may also use this method using
   *   a single identifier instead of an array.
   */
  public function load($identifiers) {
    // Single value given.
    if (!is_array($identifiers)) {
      $identifiers = array($identifiers);
    }
    // Avoid duplicates.
    else {
      $identifiers = array_unique($identifiers);
    }

    $objects = array();

    // Fetch objects. This will ensure we will get through cache.
    foreach ($identifiers as $oid) {
      $objects[] = $this->loadByIdentifier($oid);
    }

    return $objects;
  }

  /**
   * Tells if a field exists in schema API for the current object type.
   * 
   * @return boolean
   */
  protected final function _fieldExists($field) {
    $schema = drupal_get_schema($this->getTableName());
    return !empty($schema) && isset($schema['fields'][$field]);
  }

  /**
   * This will create the query conditions that matches the given access
   * permission you can then use as the $condition parameter for load
   * methods.
   * 
   * @param string $op
   *   'delete', 'update' or 'view'. Object creation here as no sens because
   *   only existing objects are going to be loaded from database.
   * @param int|object $account = NULL
   *   (optional) Account identifier or object. If NULL it will uses the
   *   current user for access checks.
   * 
   * @return array
   *   Key/value pairs of parameters. Keys are field name, and values are
   *   values to match in the query.
   *   Returns FALSE in case an invalid permission is given, or if the user
   *   has no right at all.
   */
  protected final function _getUserAccessConditions($op, $account = NULL) {
    // Check if an account was given. We need the full object structure
    // for the user_access method.
    if (!$account) {
      global $user;
      $account = $user;
    }
    else {
      $account = is_object($account) ? $account : user_load($account);
    }
    // Do the proper object exclusions following the permission.
    $perm_any = $op . ' any ' . $this->_type;
    $perm_own = $op . ' own ' . $this->_type;
    // Check user access.
    if (user_access($perm_any)) {
      // No filtering when 'any' type permission is given.
      return array();
    }
    else if (user_access($perm_own)) {
      return array('uid' => $account->uid);
    }
    else {
      return FALSE;
    }
  }

  /**
   * Get base query for given factory object type. The returned query will be
   * populated with current user access conditions.
   * 
   * This query then can be conditionned as needed, either by arbitrary joins
   * either by using any condition on the given table.
   * 
   * @param string $alias = 'ct'
   *   (optional) Alias to use for this factory table.
   * @param string $perm = 'view'
   *   (optional) Which perm to use for query. Perm can be either own of those:
   *     - 'view'
   *     - 'update'
   *     - 'delete'
   *     - NULL: Will remove any permission check. Do not never use this for
   *       any end-user oriented query. 
   * @param boolean $excludeDisabled = TRUE
   *   (optional) If set to FALSE, disabled object will be load too. This is
   *   useful for administrator forms or administration pages.
   * 
   * @return SelectQuery
   *   SelectQuery.
   */
  public function getQuery($alias = 'ct', $perm = 'view', $excludeDisabled = TRUE) {
    // Prepare the query.
    $select = db_select($this->getTableName(), $alias)
      ->fields($alias, array('oid'));

    if ($perm) {
      $access_conditions = $this->_getUserAccessConditions($perm);
      // If user has no rights at all, put a trivial and always false condition
      // to avoid any rows being returned.
      // FIXME: Needs testing.
      if (!is_array($access_conditions)) {
        $select->condition(1, 0, '=');
      }
      else {
        foreach ($access_conditions as $field => $values) {
          $select->condition($field, $values, (is_array($values) ? 'IN' : '='));
        }
      }
    }

    // Check for status exclusion.
    if ($excludeDisabled) {
      $select->condition('status', XoxoObject::STATUS_DISABLED, '<>');
    }

    return $select;
  }

  /**
   * Save a layout instance into database. This will not save included
   * content.
   * 
   * @param XoxoExportableObject $object
   * 
   * @throws XoxoFactoryException
   *   In case of invalid instance given.
   */
  public final function save(XoxoObject $object) {
    $insert = FALSE;

    // Get a clean object clone, and set data for database saving.
    $prepared = clone($object);
    $prepared->clean();
    $data = array(
      // This field won't be used anymore.
      // 'data' => serialize($prepared),
      'uid' => $object->getOwnerUid(),
      'status' => $object->getStatus(),
      'type' => $object->getType(),
    );

    // Try to get back current identifier, if none exists, then we are doing an
    // insert.
    try {
      $oid = $object->getIdentifier();
      $update = array('oid');
      $data['oid'] = $oid;
      $data['updated'] = time();
      // An updated object is always a non default.
      $data['defaults'] = 0;
      $ret = self::SAVED_UPDATED;
    }
    catch (OoxException $e) {
      $insert = TRUE;
      $update = array();
      $data['updated'] = $data['created'] = time();
      $data['defaults'] = $object->isDefault() ? 1 : 0;
      $ret = self::SAVED_NEW;
    }

    // Create a transaction (this won't work with MyISAM) to ensure we can set
    // the name property to the oid property.
    // We do this because when operating on XoxoObject, name has no sense
    // name must be used only for exportables that needs a unique machine name.
    // Note that in any case, there really a few chances that this transaction
    // overlaps with another, so, if you are lucky, it may still work using the
    // MyISAM MySQL engine.
    $transaction = db_transaction();

    // Last chance modification for overriding classes.
    $this->_preSave($object, $data);

    try {
      // Real save.
      if (drupal_write_record($this->getTableName(), $data, $update) === FALSE) {
        $ret = self::SAVED_ERROR;
      }
      else {
        // This should always goes here.
        if (empty($data['name'])) {
          db_update($this->getTableName())
            ->fields(array('name' => $data['oid']))
            ->condition('oid', $data['oid'])
            ->execute();
        }
        $this->_postSave($object, $data);
      }

      // If no error, and we are in the case of object creation, update cache.
      // In case of object update, we don't need it because object reference is
      // already in the static cache. Clones are BAD, and should never be done
      // by other modules.
      if ($insert) {
        $object->setIdentifier($data['oid']);
      }
      else {
        // Reset default flag at update.
        $object->setDefault(FALSE);
      }

      // Update the cache if needed.
      $this->_cacheAdd($object, TRUE);
    }
    catch (OoxException $e) {
      watchdog_exception('xoxo_factory', $e);
      $ret = self::SAVED_ERROR;
      $transaction->rollback();
    }
    catch (PDOException $e) {
      watchdog_exception('xoxo_factory', $e);
      $ret = self::SAVED_ERROR;
      $transaction->rollback();
    }

    return $ret;
  }

  /**
   * Last chance for overriding classes to alter the $data array when saving.
   * This is an internal method, please do not override it unless you really
   * know what you are doing.
   * 
   * This function is run within a database transaction, any request that
   * does an error here will then implies a database rollback and object
   * save failure.
   * 
   * @param XoxoObject $object
   *   Object being saved. If this is an insert, the object does not have
   *   its identifier set yet.
   * @param array &$data
   *   The data array that will be passed to drupal_write_record().
   */
  protected function _preSave(XoxoObject $object, &$data) { }

  /**
   * Save has been done, with no errors.
   * 
   * This function is run within a database transaction, any request that
   * does an error here will then implies a database rollback and object
   * save failure.
   * 
   * @param XoxoObject $object
   *   Object being saved. If this is an insert, the object does not have
   *   its identifier set yet.
   * @param array &$data
   *   The data array that got through to drupal_write_record(), identifier
   *   is set in this array as $data['oid'].
   */
  protected function _postSave(XoxoObject $object, &$data) {
    // Save all options. We need to do this in postSave else we won't have any
    // object oid in case of object creation.
    $optionsOid = $object->getOptionsIdentifier();
    $oid = $object->getOptionsIdentifier();
    OptionsFactory::getInstance()->save($object);
    // We save options, we need to alter the existing row.
    if ($optionsOid != ($newOid = $oid)) {
      db_update($this->getTableName())
        ->fields(array('options_oid' => $newOid))
        ->condition('oid', $oid)
        ->execute();
    }
  }

  /**
   * Delete object.
   * 
   * Using the identifier to delete an object is mandatory, because other
   * factories may want to delete other stuff arround. Delete has no error
   * handling, once the object has been deleted, there is no comming back.
   * 
   * @see XoxoFactory::_postDelete()
   * 
   * @param int $oid
   *   XoxoObject identifier.
   * 
   * @return int
   *   Number of rows deleted.
   */
  public final function deleteByIdentifier($oid) {
    return $this->deleteByField(array('oid' => $oid));
  }

  /**
   * Every factory that uses secondary database tables to store data arround
   * the objects should use this method to clean their data. Use this method
   * to also cleanup your caches.
   * 
   * Their is no transaction here and object is going to be delete, this is
   * your last chance to delete all remaining garbage.
   * 
   * @see XoxoFactory::deleteByIdentifier()
   * 
   * @param XoxoObject $object
   *   XoxoObject identifier.
   */
  protected function _preDelete(XoxoObject $object) {
    // Removes all outdated data.
    if ($optionsOid = $object->getOptionsIdentifier()) {
      OptionsFactory::getInstance()->delete($optionsOid);
    }
  }

  /**
   * Delete objects using the given table field. This field must exist in
   * current table schema API cache.
   * 
   * It will do two requests, the first one will fetch back identifiers so
   * we can erase only a part of the cache, the second one will really delete
   * those objects using a IN statement for identifiers, to ensure no database
   * update have been done between our two requests.
   * 
   * @param array $conditions
   *   Array of conditions, each value is a field name, each value is a field
   *   value.
   * 
   * @return int
   *   Number of rows deleted.
   */
  public function deleteByField($conditions) {
    $deleted = 0;
    $table = $this->getTableName();

    $transaction = db_transaction();

    try {
      // First request.
      $select = db_select($table, 'ct')
        ->fields('ct', array('oid'));
      foreach ($conditions as $field => $value) {
        $select->condition($field, $value);
      }
      $oids = $select->execute()->fetchCol();

      // Proceed only if we have something to delete.
      if ($deleted = count($oids)) {
        // Execute post delete hook, this is important.
        foreach ($oids as $oid) {
          // FIXME: This is not good, loading an object in order to delete it.
          // But at least, it gives a chance for overriding modules to have it
          // for latest cleaning operations.
          $this->_preDelete($this->loadByIdentifier($oid));
        }

        // Second query.
        db_delete($table)
          ->condition('oid', $oids, 'IN')
          ->execute();

        // Force transaction to commit.
        unset($transaction);

        // Then also clear nicely the cache.
        $this->cacheClear($oids);
      }
    }
    catch (PDOException $e) {
      // This should not happen.
      watchdog_exception('xoxo_factory', $e);
    }

    return $deleted;
  }

  /**
   * Delete objects owned by the given account.
   * 
   * @param int|object $account
   *   Account identifier or object.
   * 
   * @return int
   *   Number of rows deleted.
   */
  public function deleteByOwner($account) {
    $uid = is_object($account) ? $account ->uid : $account;
    return $this->deleteByField(array('uid', $uid));
  }

  /**
   * Check if the given identifier exists in database.
   * 
   * @param int $oid
   *   Identifier.
   * 
   * @return boolean
   */
  public function existsByIdentifier($oid) {
    return isset($this->_loaded[$oid]) || (bool) db_result(db_query("SELECT 1 FROM {:table} WHERE oid = :oid", array(':table' => $this->getTableName(), ':oid' => $oid)));
  }

  /**
   * Factory method used by common code. This should be overriden by any
   * specific factory implementation.
   * 
   * @param string $type
   *   Registry item type. If given, will spawn a factory named using the
   *   factory type as registry type and aptempt to create the given item
   *   type as class to instanciate the object.
   *
   * @return XoxoObject
   *   New empty instance.
   * 
   * @throws OoxRegistryException
   *   When item is not supported by the registry itself.
   * @throws XoxoFactoryException
   *   When type is not supported.
   */
  public final function createInstance($type) {
    if ($this->isItemTypeSupported($type)) {
      return $this->getItemRegistry()->getItem($type);
    }
    throw new XoxoFactoryException("Unsupported item type " . $type);
  }

  /**
   * Change only object status in database, leaving all other stuff unchanged.
   * Status must have already been changed in object.
   * 
   * @param XoxoObject $object
   */
  public function saveStatus(XoxoObject $object) {
    db_update($this->getTableName())
      ->fields(array('status' => $object->getStatus()))
      ->condition('oid', $object->getIdentifier())
      ->execute();
  }

  /**
   * Add a loaded object to cache.
   * 
   * For overriding classes, this is the right method to override to add your
   * own cache mecanism and implementation.
   * 
   * If you override it, please call the parent method or update carefully
   * the protected _loaded property, or call the parent implementation into
   * your own.
   * 
   * @param XoxoObject $object
   *   Object to add to cache.
   * @param boolean $toBinary = FALSE
   *   (optional) If set to TRUE, cache this object into binary cache.
   *   Implementors that override this function may use their own binary
   *   cache table.
   * 
   * @param XoxoObject $object
   *   Object to add to static cache.
   */
  public function _cacheAdd(XoxoObject $object, $toBinary = FALSE) {
    $oid = $object->getIdentifier();

    // Static cache.
    $this->_loaded[$oid] = $object;

    // Binary cache.
    if ($toBinary) {
      $cid = $this->getTableName() . '-' . $oid;
      cache_set($cid, $object, XOXO_CACHE_OBJECT_TABLE);
    }
  }

  /**
   * Get item from binary cache.
   * 
   * Basic implementation will use only the first parameter as oid. Implementors
   * may use as many parameters as they want, to build more complex cache id
   * (exportable objects may embed the name parameter as well).
   * 
   * @see XoxoExportableFactory::_cacheGetBinary()
   *   For a sample implementation.
   */
  protected function _cacheGetBinary() {
    $args = func_get_args();

    $oid = $args[0];
    $cid = $this->getTableName() . '-' . $oid;

    if ($data = cache_get($cid, XOXO_CACHE_OBJECT_TABLE)) {
      return $data->data;
    }
  }

  /**
   * Clear internal static cache.
   * 
   * @param array $oids = array()
   *   (optional) Objects to remove from static cache.
   */
  public final function cacheClear($oids = array()) {
    $this->_cacheClear($oids);
  }

  /**
   * Clear internal static cache.
   * 
   * The first parameter will allow the method to clear only given objects from
   * static cache. Any overriding class may override this method to ensure that
   * all caches are cleared. When implementing it, if you sense that clearing
   * correctly your caches could take more time than reloading the objects, then
   * erase it all. Remember it's better to clean it completely than let a wrong
   * static cache living.
   * 
   * If you override it, please call the parent method or update carefully
   * the protected _loaded property.
   * 
   * @param array $oids = array()
   *   (optional) Objects to remove from static cache.
   */
  protected function _cacheClear($oids = array()) {
    if (!empty($oids)) {
      foreach ($oids as $oid) {
        // See below notice for living instances.
        unset($this->_loaded[$oid]);
      }
    }
    else {
      // This will keep out of scope every objects in the cache, so that the
      // __destruct() magic method will be called on them. Of course, this will
      // work only if the objects have no left references currently being used
      // by custom modules.
      $this->_loaded = array();
    }
  }

  /**
   * Get help text that will display on administration overview page.
   * 
   * FIXME: Quite ugly, find a better pattern than this.
   * 
   * @return string
   *   Human readable localized string.
   */
  public function getOverviewHelp() { return ""; }
}
